#ifndef ROOTCNT_H
#define ROOTCNT_H

#include "romberg.hpp" // used for numerical integration
#include <cmath>

#ifndef M_PI
#define M_PI           3.14159265358979323846
#endif

/*
   Compute number of simple roots of a function f: [a,b] --> |R and a,b in |R
   where
     (1) f is twice continuously differentiable in [a,b]
     (2) f(a) * f(b) != 0

   The computation is done using a Kronecker-Picard integral,
   see V.P. Plagianakos et al: "Locating and computing in parallel
   all the simple roots of special functions using PVM",
   Journal of Computational and Applied Mathematics 133 (2001) 545-554

   Parameters:
      F     function f
      F1    function f'
      F2    function f''
      Real  type which must support atan, fabs, and regular arithmetic ops
            (by default double)

   Example:
      struct F { double operator()(double x) const { return sin(x); } };
      struct F1 { double operator()(double x) const { return cos(x); } };
      struct F2 { double operator()(double x) const { return -sin(x); } };
      F f; F1 f1; F2 f2;
      RootCount<F, F1, F2> rootcount(f, f1, f2, -1, 8);
      unsigned int count = rootcount; // delivers 3
*/

template<typename Real = double>
class RootCount {
   public:
      template<typename F, typename F1, typename F2>
      class Integrand {
	 public:
	    Integrand(F& f_param, F1& f1_param, F2& f2_param) :
	       f(f_param), f1(f1_param), f2(f2_param) {
	    }
	    Real operator()(Real x) const {
	       Real fx = f(x);
	       Real f1x = f1(x);
	       Real f1xsq = f1x * f1x;
	       return (fx * f2(x) - f1xsq) / (fx * fx + f1xsq);
	    }
	 private:
	    F& f; F1& f1; F2& f2;
      };
      template<typename F, typename F1, typename F2>
      unsigned int get_count(F& f, F1& f1, F2& f2,
            Real a, Real b) {
	 Integrand<F, F1, F2> integrand(f, f1, f2);
	 Romberg<Real> integral(integrand, a, b, 1E-2);
	 Real n = - (integral - std::atan(f1(b)/f(b)) +
			std::atan(f1(a)/f(a))) / M_PI;
	 return (unsigned int) (n + 0.1);
      }
};

#endif
